/* © SRSoftware 2024 */
package de.srsoftware.cal.api;

import java.time.LocalDateTime;
import java.util.Optional;
import java.util.Set;
import org.json.JSONObject;

/**
 * This is the central object of the calendar: An appointment
 */
public interface Appointment {
	/**
	 * list of attachments associated with the appointment. may be used for flyers or images
	 * @return list of attachments
	 */
	Set<Attachment> attachments();

	/**
	 * defines the location of the event
	 * @return optional location
	 */
	Optional<Coords> coords();

	Appointment clone(long newId);

	/**
	 * Descriptive text of the event
	 * @return the description
	 */
	String description();

	/**
	 * The date and time when the appointment is scheduled to end
	 * @return an optionals LocalDateTime
	 */
	Optional<LocalDateTime> end();

	/**
	 * represent this event as ical entry
	 * @param calendarIdentifier used to generate a unique id. pass something that is unique, but stable over time, e.g. a hostname
	 * @param urlTemplate the url to put into the ical, with "{}" being replaced by the appointment id
	 * @return the ical string
	 */
	String ical(String calendarIdentifier,String urlTemplate);

	/**
	 * ID of the appointment – unique within this system
	 * @return the appointment`s id
	 */
	long id();

	/**
	 * get a json representation of this Appointment
	 * @param urlTemplate the url to put into the ical, with "{}" being replaced by the appointment id
	 * @return a JSON Object
	 */
	JSONObject json(String urlTemplate);

	/**
	 * set of Links that point to related information
	 * @return set of links
	 */
	Set<Link> links();

	/**
	 * descriptive text of the location, e.g. address
	 * @return location text
	 */
	Optional<String> location();

	/**
	 * The date and time, when the appointment starts
	 * @return the start time as LocalDateTime
	 */
	LocalDateTime start();

	/**
	 * tags i.e. keywords may be used to filter appointments
	 * @return the set of tags associated with the current appointment
	 */
	Set<String> tags();

	/**
	 * the title of the appointment
	 * @return the title
	 */
	String title();
}
