/* © SRSoftware 2024 */
package de.srsoftware.cal.app;

import static java.lang.System.Logger.Level.*;

import com.sun.net.httpserver.HttpServer;
import de.srsoftware.cal.*;
import de.srsoftware.cal.db.Database;
import de.srsoftware.cal.db.MariaDB;
import de.srsoftware.configuration.Configuration;
import de.srsoftware.configuration.JsonConfig;
import de.srsoftware.tools.ColorLogger;
import de.srsoftware.tools.plugin.JarWatchdog;
import java.io.File;
import java.io.IOException;
import java.net.InetSocketAddress;
import java.sql.SQLException;
import java.time.Duration;
import java.util.Optional;
import org.json.JSONArray;
import org.json.JSONObject;

/**
 * Test application
 */
public class Application {
	private static final System.Logger LOG     = System.getLogger(Application.class.getSimpleName());
	private static final String	   JDBC    = "opencloudcal.db.jdbc";
	private static final String	   USER    = "opencloudcal.db.user";
	private static final String	   PASS    = "opencloudcal.db.pass";
	private static final String	   MISSING = "missing required configuration property \"%s\"";

	private Application() {
	}

	private static Database connect(Configuration config) throws SQLException {
		Optional<String> jdbc = config.get(JDBC);
		if (jdbc.isEmpty()) throw new RuntimeException(MISSING.formatted(JDBC));

		String user = config.get(USER, "opencloudcal");

		Optional<String> pass = config.get(PASS);
		if (pass.isEmpty()) throw new RuntimeException(MISSING.formatted(PASS));

		return MariaDB.connect(jdbc.get(), user, pass.get());
	}

	/**
	 * sandbox
	 * @param args default
	 */
	public static void main(String[] args) throws IOException, SQLException {
		LOG.log(INFO, "Starting application…");
		JsonConfig	 jsonConfig  = new JsonConfig("OpenCloudCal");

		configureLogging(jsonConfig);

		Optional<String> staticPath  = jsonConfig.get("opencloudcal.static");
		var	 db          = connect(jsonConfig);
		var	 port        = jsonConfig.get("opencloudcal.http.port", 8080);
		HttpServer	 server      = HttpServer.create(new InetSocketAddress(port), 0);
		var	 staticPages = new StaticHandler(staticPath).bindPath("/static").on(server);
		new IndexHandler(staticPages).bindPath("/").on(server);
		new ApiEndpoint(db).bindPath("/api").on(server);
		new Forward().bindPath("/index.php").on(server);

		server.start();
		LOG.log(INFO,"Started webserver on port {0,number,#} …",port);

		var autoImport = new AutoImporter(db);
		var watchDog = new JarWatchdog()
				.setContext(BaseImporter.class.getClassLoader())
				.frequency(Duration.ofSeconds(30))
				.addListener(autoImport)
				.afterScan(autoImport);

		var importers = jsonConfig.get("opencloudcal.importers").orElse(null);
		if (importers instanceof JSONArray arr) arr.toList().stream()
				.map(Object::toString).map(File::new).forEach(watchDog::addDirectory);
		watchDog.start();
	}

	private static void configureLogging(JsonConfig jsonConfig) {
		var rootLevel = jsonConfig.get("opencloudcal.logging.rootLevel", "INFO");
		ColorLogger.setRootLogLevel(mapLogLevel(rootLevel));

		var loggers = jsonConfig.get("opencloudcal.logging.loggers").orElse(null);
		if (loggers instanceof JSONObject json){
			json.keySet().forEach(key -> {
				var lvl = mapLogLevel(json.getString(key));
				ColorLogger.setLogLevel(key,lvl);
			});
		}
	}

	private static System.Logger.Level mapLogLevel(String lbl) {
		return switch (lbl){
			case "ERROR" -> ERROR;
			case "WARN" -> WARNING;
			case "DEBUG" -> DEBUG;
			case "TRACE" -> TRACE;
			default -> INFO;
		};
	}

}
