/* © SRSoftware 2024 */
package de.srsoftware.cal;

import static de.srsoftware.cal.Util.*;
import static de.srsoftware.tools.Optionals.nullable;

import de.srsoftware.cal.api.Appointment;
import de.srsoftware.cal.api.Attachment;
import de.srsoftware.cal.api.Coords;
import de.srsoftware.cal.api.Link;
import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.util.*;
import org.json.JSONObject;

/**
 * basic class for Appointments
 */
public class BaseAppointment implements Appointment {
	private static final DateTimeFormatter DATE_TIME = DateTimeFormatter.ofPattern("yyyy-MM-dd HH:mm");
	private long		       id;
	private final String	       title, description;
	private final LocalDateTime	       end, start;
	private Coords		       coords = null;
	private final Set<Attachment> attachments     = new HashSet<>();
	private final Set<String> tags	              = new HashSet<>();
	private final Set<Link> links	              = new HashSet<>();
	private final String    location;

	/**
	 * create a new appointment
	 * @param id set the id
	 * @param title set the title
	 * @param description set the description
	 * @param start set the start date
	 * @param end set the end date
	 * @param location set the location
	 */
	public BaseAppointment(long id, String title, String description, LocalDateTime start, LocalDateTime end, String location) {
		this.description = description;
		this.end	 = end;
		this.id	 = id;
		this.location	 = location;
		this.start	 = start;
		this.title	 = title;
	}

	/**
	 * adds attachments
	 * @param newAttachments the attachments to add to the appointment
	 * @return the appointment
	 */
	public BaseAppointment add(Attachment... newAttachments) {
		Collections.addAll(attachments, newAttachments);
		return this;
	}

	/**
	 * adds attachments
	 * @param newAttachments the attachments to add to the appointment
	 * @return the appointment
	 */
	public BaseAppointment add(Collection<Attachment> newAttachments) {
		attachments.addAll(newAttachments);
		return this;
	}

	/**
	 * adds links
	 * @param newLinks the links to add to the appointment
	 * @return the appointment
	 */
	public BaseAppointment addLinks(Link... newLinks) {
		Collections.addAll(links, newLinks);
		return this;
	}

	/**
	 * adds links
	 * @param newLinks the links to add to the appointment
	 * @return the appointment
	 */
	public BaseAppointment addLinks(Collection<Link> newLinks) {
		links.addAll(newLinks);
		return this;
	}

	/**
	 * set the coordinates of the attachments
	 * @param newCoords the coordinates to apply
	 * @return the appointment
	 */
	public BaseAppointment coords(Coords newCoords) {
		coords = newCoords;
		return this;
	}

	@Override
	public Set<Attachment> attachments() {
		return attachments;
	}

	@Override
	public Appointment clone(long newId) {
		return new BaseAppointment(newId, title, description, start, end, location)  //
		    .coords(coords)
		    .addLinks(links)
		    .add(attachments)
		    .tags(tags);
	}

	@Override
	public Optional<Coords> coords() {
		return nullable(coords);
	}

	@Override
	public String description() {
		return description;
	}

	@Override
	public Optional<LocalDateTime> end() {
		return nullable(end);
	}

	@Override
	public String ical(String calendarIdentifier,String urlTemplate) { // TODO: implement
		var sb = new StringBuilder();
		sb.append(contentLine(BEGIN,VEVENT));
		if (calendarIdentifier != null) sb.append(contentLine(UID,"%s@%s".formatted(id(),calendarIdentifier)));
		sb.append(contentLine(DTSTART,start().format(ICAL_DATE_FORMAT).replace(" ","T")));
		end().map(end -> contentLine(DTEND,end.format(ICAL_DATE_FORMAT).replace(" ","T"))).ifPresent(sb::append);
		sb.append(contentLine(SUMMARY,title()));
		sb.append(contentLine(DESCRIPTION,description()));
		sb.append(contentLine(URL,urlTemplate.replace("{}",""+id())));
		coords().map(Coords::icalFormat).map(geo -> contentLine(GEO,geo)).ifPresent(sb::append);
		location().map(loc -> contentLine(LOCATION,loc)).ifPresent(sb::append);
		for (var attachment : attachments()) sb.append(contentLine("ATTACH;FMTYPE=%s".formatted(attachment.mime()),attachment.url().toString()));
		for (var link : links) sb.append(contentLine("ATTACH;TITLE="+paramText(link.desciption()),link.url().toString()));
		sb.append(contentLine("CLASS","PUBLIC"));
		sb.append(contentLine(END,VEVENT));
		return sb.toString();
	}


	@Override
	public long id() {
		return id;
	}

	@Override
	public JSONObject json(String urlTemplate) {
		var json = new JSONObject();
		json.put("attachments", attachments.stream().map(Attachment::json).toList());
		json.put("coords", nullable(coords).map(Coords::json).orElse(null));
		json.put("description", description());
		json.put("end", end().map(end -> end.format(DATE_TIME)).orElse(null));
		json.put("id", id());
		json.put("location", location().orElse(null));
		json.put("start", start().format(DATE_TIME));
		json.put("tags", tags());
		json.put("title", title());
		json.put("links", links().stream().map(Link::json).toList());
		json.put("url",urlTemplate.replace("{}",""+id()));
		return json;
	}

	@Override
	public Optional<String> location() {
		return nullable(location);
	}

	@Override
	public LocalDateTime start() {
		return start;
	}

	@Override
	public Set<String> tags() {
		return tags;
	}

	/**
	 * adds tag
	 * @param newTags the tag to add to the appointment
	 * @return the appointment
	 */
	public BaseAppointment tags(String... newTags) {
		Collections.addAll(tags, newTags);
		return this;
	}

	/**
	 * adds tag
	 * @param newTags the tag to add to the appointment
	 * @return the appointment
	 */
	public BaseAppointment tags(Collection<String> newTags) {
		tags.addAll(newTags);
		return this;
	}

	@Override
	public String title() {
		return title;
	}

	@Override
	public String toString() {
		return "%s (%s)".formatted(title, start);
	}

	@Override
	public Set<Link> links() {
		return links;
	}
}
