/* © SRSoftware 2024 */
package de.srsoftware.cal;

import static de.srsoftware.cal.Util.*;
import static de.srsoftware.tools.Error.error;
import static de.srsoftware.tools.Result.transform;
import static de.srsoftware.tools.Tag.HREF;
import static de.srsoftware.tools.TagFilter.*;

import de.srsoftware.cal.api.*;
import de.srsoftware.tools.*;
import java.io.InputStream;
import java.net.*;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.LocalTime;
import java.util.List;
import java.util.Objects;
import java.util.Optional;
import java.util.function.Predicate;
import java.util.stream.Stream;

public abstract class BaseImporter implements Importer {
	public static final System.Logger LOG = System.getLogger(BaseImporter.class.getSimpleName());
	private static final String	   SHA256 = "SHA-256";

	protected abstract String baseUrl();

	@Override
	public abstract String description();

	protected List<Attachment> extractAttachments(Tag eventTag) {
		return extractAttachmentsTag(eventTag)  //
				.optional()
				.stream()
				.flatMap(tag -> tag.find(IS_IMAGE).stream())
				.map(tag -> tag.get("src"))
				.filter(Objects::nonNull)
				.map(url -> url.contains("://") ? url : baseUrl()+url)
				.map(Payload::of)
				.map(Util::url)
				.map(Util::toAttachment)
				.map(Result::optional)
				.flatMap(Optional::stream)
				.toList();
	}

	protected abstract Predicate<Tag> extractAttachmentsFilter();

	protected Result<Tag> extractAttachmentsTag(Tag eventTag) {
		var list = eventTag.find(extractAttachmentsFilter());
		if (list.isEmpty()) return error("Failed to find attachments tag");
		return Payload.of(list.getFirst());
	}


	protected Result<String> extractDescription(Tag eventTag) {
		Result<Tag> descriptionTag = extractDescriptionTag(eventTag);
		if (descriptionTag.optional().isEmpty()) return transform(descriptionTag);
		Tag tag = descriptionTag.optional().get();
		tag.find(ofType("iframe")).forEach(Tag::remove);
		var inner = tag.inner(2);
		return inner.isPresent() ? Payload.of(inner.get()) : error("No description found");
	}

	protected abstract Predicate<Tag> extractDescriptionFilter();

	protected Result<Tag> extractDescriptionTag(Tag eventTag){
		var list = eventTag.find(extractDescriptionFilter());
		if (list.isEmpty()) return error("Failed to find description tag");
		return Payload.of(list.getFirst());
	}


	protected abstract Result<Coords> extractCoords(Tag eventTag);

	protected Result<LocalDateTime> extractEnd(Tag eventTag) {
		Result<LocalDate> date = extractEndDate(eventTag);
		Result<LocalTime> time = extractEndTime(eventTag);
		return combine(date,time);
	}

	protected Result<LocalDate> extractEndDate(Tag eventTag) {
		Result<Tag> endDateTag = extractEndDateTag(eventTag);
		var opt = endDateTag.optional();
		if (opt.isEmpty()) return transform(endDateTag);
		return parseEndDate(opt.get().strip());
	}

	protected Result<Tag> extractEndDateTag(Tag eventTag) {
		var list = eventTag.find(extractEndDateFilter());
		if (list.isEmpty()) return error("Failed to find end date tag");
		return Payload.of(list.getFirst());
	}

	protected abstract Predicate<Tag> extractEndDateFilter();

	protected Result<LocalTime> extractEndTime(Tag eventTag) {
		Result<Tag> endTimeTag = extractEndTimeTag(eventTag);
		var opt = endTimeTag.optional();
		if (opt.isEmpty()) return transform(endTimeTag);
		return parseEndTime(opt.get().strip());
	}

	protected Result<Tag> extractEndTimeTag(Tag eventTag) {
		var list = eventTag.find(extractEndTimeFilter());
		if (list.isEmpty()) return error("Failed to find end time tag");
		return Payload.of(list.getFirst());
	}

	protected abstract Predicate<Tag> extractEndTimeFilter();


	protected Result<Appointment> extractEvent(Result<Tag> domResult, Link eventPage) {
		var  opt = domResult.optional();
		if (opt.isEmpty()) return transform(domResult);
		var eventTag = opt.get();

		long id = 0;

		// wird vor extractDescription ausgeführt, da extractDescription das DOM verändert
		var coords = extractCoords(eventTag);

		var titleResult = extractTitle(eventTag);
		if (titleResult.optional().isEmpty()) return transform(titleResult);
		String title = titleResult.optional().get();

		var descriptionResult = extractDescription(eventTag);
		if (descriptionResult.optional().isEmpty()) return transform(descriptionResult);
		var description = descriptionResult.optional().get();

		var startResult = extractStart(eventTag);
		if (startResult.optional().isEmpty()) return transform(startResult);
		var start = startResult.optional().get();

		var endResult = extractEnd(eventTag);
		var end       = endResult.optional().orElse(null);

		var locationResult = extractLocation(eventTag);
		if (locationResult.optional().isEmpty()) return transform(locationResult);
		var location = locationResult.optional().get();

		var event = new BaseAppointment(id, title, description, start, end, location)  //
			.add(extractAttachments(eventTag))
			.addLinks(extractLinks(eventTag))
			.addLinks(eventPage)
			.tags(extractTags(eventTag));

		coords.optional().ifPresent(event::coords);

		return Payload.of(event);
	}

	/**
	 * Die ist der Tag auf der Seite, der alle weiteren Event-Daten umfasst.
	 * Im Prinzip kann der Page-Tag auch direkt weitergereicht werden, dann
	 * sind die weiteren Suchen aber Umfangreicher.
	 * Besser ist es daher, den Tag näher einzugrenzen
	 * @param pageResult
	 * @return
	 */
    protected Result<Tag> extractEventTag(Result<Tag> pageResult){
		var opt = pageResult.optional();
		if (opt.isEmpty()) return transform(pageResult);
		var list = opt.get().find(extractEventTagFilter());
		if (list.isEmpty()) return error("Failed to find event tag");
		return Payload.of(list.getFirst());
	}

	protected abstract Predicate<Tag> extractEventTagFilter();


	protected abstract Result<List<String>> extractEventUrls(Result<Tag> programPage);

	protected List<Link> extractLinks(Tag appointmentTag) {
		var tag = extractLinksTag(appointmentTag);
		var opt = tag.optional();
		if (opt.isEmpty()) return List.of();
		Tag linksTag = opt.get();
		return linksTag.find(IS_ANCHOR).stream()
				.map(anchor -> {
					var href = anchor.get(HREF);
					if (href == null) return null;
					if (!href.contains("://")) href = baseUrl()+href;
					var txt = anchor.strip().trim();
					return url(Payload.of(href)).optional().map(url -> new Link(url,txt)).orElse(null);
				})
				.filter(Objects::nonNull)
				.filter(link -> link.desciption() != null)
				.toList();
	}

    protected Result<Tag> extractLinksTag(Tag eventTag) {
		var list = eventTag.find(extractLinksFilter());
		if (list.isEmpty()) return error("Failed to find links tag");
		return Payload.of(list.getFirst());
	}

	protected abstract Predicate<Tag> extractLinksFilter();

	protected Result<String> extractLocation(Tag eventTag) {
        Result<Tag> locationTag = extractLocationTag(eventTag);
        if (locationTag.optional().isEmpty()) return transform(locationTag);
        return Payload.of(locationTag.optional().get().strip());
    }

    protected Result<Tag> extractLocationTag(Tag eventTag){
		var list = eventTag.find(extractLocationFilter());
		if (list.isEmpty()) return error("Failed to find location tag");
		return Payload.of(list.getFirst());
	}

	protected abstract Predicate<Tag> extractLocationFilter();


	protected Result<LocalDateTime> extractStart(Tag eventTag) {
		Result<LocalDate> date = extractStartDate(eventTag);
		Result<LocalTime> time = extractStartTime(eventTag);
		return combine(date,time);
	}

	protected Result<LocalDate> extractStartDate(Tag eventTag) {
		Result<Tag> startDateTag = extractStartDateTag(eventTag);
		var opt = startDateTag.optional();
		if (opt.isEmpty()) return transform(startDateTag);
		return parseStartDate(opt.get().strip());
	}

	protected Result<Tag> extractStartDateTag(Tag eventTag) {
		var list = eventTag.find(extractStartDateFilter());
		if (list.isEmpty()) return error("Failed to find start date tag");
		return Payload.of(list.getFirst());
	}

	protected abstract Predicate<Tag> extractStartDateFilter();

	protected Result<LocalTime> extractStartTime(Tag eventTag) {
		Result<Tag> startTimeTag = extractStartTimeTag(eventTag);
		var opt = startTimeTag.optional();
		if (opt.isEmpty()) return transform(startTimeTag);
		return parseStartTime(opt.get().strip());
	}

	protected Result<Tag> extractStartTimeTag(Tag eventTag) {
		var list = eventTag.find(extractStartTimeFilter());
		if (list.isEmpty()) return error("Failed to find start time tag");
		return Payload.of(list.getFirst());
	}

	protected abstract Predicate<Tag> extractStartTimeFilter();


	protected abstract List<String> extractTags(Tag eventTag);

    protected Result<String> extractTitle(Tag eventTag) {
		Result<Tag> titleTag = extractTitleTag(eventTag);
        if (titleTag.optional().isEmpty()) return transform(titleTag);
        var inner = titleTag.optional().flatMap(tag -> tag.inner(2));
        return inner.isPresent() ? Payload.of(inner.get().trim()) : error("No title found");
	}

	protected Result<Tag> extractTitleTag(Tag eventTag){
		var list = eventTag.find(extractTitleFilter());
		if (list.isEmpty()) return error("Failed to find title tag");
		return Payload.of(list.getFirst());
	}

	protected abstract Predicate<Tag> extractTitleFilter();

	@Override
	public Stream<Appointment> fetch() {
		var	       url    = Payload.of(programURL());
		Stream<Result<String>> urls = url(url)
					.map(Util::open)  //
					.map(this::preload)
					.map(Util::parseXML)
					.map(this::extractEventUrls)
					.stream();
		return urls  //
			.map(Util::url)
			.map(this::loadEvent)
			.peek(res ->  { if (res.optional().isEmpty()) LOG.log(System.Logger.Level.WARNING,res); })
			.flatMap(result -> result.optional().stream());
	}

	protected static <T> Result<T> invalidParameter(Result<?> result) {
		return error("Invalid parameter: %s", result.getClass().getSimpleName());
	}

	protected static Result<Link> link(Result<URL> url, String text) {
		var opt = url.optional();
		return opt.isEmpty() ? transform(url) : Payload.of(new Link(opt.get(), text));
	}

	protected Result<Appointment> loadEvent(Result<URL> urlResult) {
		var link = urlResult  //
					   .optional()
					   .map(url -> new Link(url, "Event-Seite"))
					   .orElse(null);
		return urlResult  //
			.map(Util::open)
			.map(this::preload)
			.map(Util::parseXML)
			.map(this::extractEventTag)
			.map(tagResult -> extractEvent(tagResult, link));
	}



	protected abstract Result<LocalDate> parseEndDate(String string);
	protected abstract Result<LocalTime> parseEndTime(String string);

	protected abstract Result<LocalDate> parseStartDate(String string);
	protected abstract Result<LocalTime> parseStartTime(String string);



    protected Result<InputStream> preload(Result<InputStream> inputStream) {
		return Util.preload(inputStream);
	}

	protected abstract String programURL();

}
