/* © SRSoftware 2024 */
package de.srsoftware.cal.importer.jena;

import static de.srsoftware.cal.Util.parseGermanDate;
import static de.srsoftware.cal.Util.parseGermanTime;
import static de.srsoftware.tools.Error.error;
import static de.srsoftware.tools.Result.transform;
import static de.srsoftware.tools.Tag.CLASS;
import static de.srsoftware.tools.TagFilter.*;
import static java.nio.charset.StandardCharsets.UTF_8;

import de.srsoftware.cal.BaseImporter;
import de.srsoftware.cal.api.Attachment;
import de.srsoftware.cal.api.Coords;
import de.srsoftware.tools.*;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.security.NoSuchAlgorithmException;
import java.time.LocalDate;
import java.time.LocalTime;
import java.util.List;
import java.util.function.Predicate;
import java.util.regex.Pattern;

public class CosmicDawn extends BaseImporter {
	private static final Pattern START_DATE_PATTERN = Pattern.compile("(\\d\\d?).(\\d\\d?).(\\d{4}).*(\\d\\d?):(\\d\\d?)");
private static final String DEFAULT_LOCATION = "Cosmic Dawn e.V., Spitzweidenweg 28, 07743 Jena";
	private static final Coords DEFAULT_COORDS = new Coords(50.93663, 11.59254);

	public CosmicDawn() throws NoSuchAlgorithmException {
		super();
	}

	@Override
	protected String baseUrl() {
		return "https://www.kuba-jena.de";
	}

	@Override
	public String description() {
		return "Importiert Events des Kulturbahnhofs in Jena";
	}

	@Override
	protected List<Attachment> extractAttachments(Tag eventTag) {
		return super.extractAttachments(eventTag).stream()
				.filter(att -> !att.url().toString().contains("cropped-logo_square"))
				.toList();
	}

	@Override
	protected Predicate<Tag> extractAttachmentsFilter() {
		return attributeContains(CLASS,"single-event-page");
	}

	@Override
	protected Predicate<Tag> extractDescriptionFilter() {
		return attributeContains(CLASS,"event-body-content");
	}

	@Override
	protected Result<Coords> extractCoords(Tag eventTag) {
		return Payload.of(DEFAULT_COORDS);
	}

	@Override
	protected Predicate<Tag> extractEndDateFilter() {
		return null;
	}

	@Override
	protected Predicate<Tag> extractEndTimeFilter() {
		return null;
	}

	protected Result<Tag> extractEventTag(Result<Tag> pageResult) {
		var res = super.extractEventTag(pageResult);
		// remove youtube embeddings
		if (res instanceof Payload<Tag> payload) payload.get().find(attributeContains(CLASS,"youtube")).forEach(Tag::remove);
		return res;
	}

		@Override
	protected Predicate<Tag> extractEventTagFilter() {
		return attributeEquals(CLASS,"inside-article");
	}

	@Override
	protected Result<List<String>> extractEventUrls(Result<Tag> programPage) {
		var page = programPage.optional();
		if (page.isEmpty()) return transform(programPage);
		var list = page.get().find(attributeEquals("class","event_listings_main"));
		var urlList = list.stream()
				.flatMap(tag -> tag.find(IS_ANCHOR).stream())
				.map(tag -> tag.get("href"))
				.toList();
		return Payload.of(urlList);
	}

	@Override
	protected Predicate<Tag> extractLinksFilter() {
		return attributeContains(CLASS,"single-event-wrapper");
	}

	@Override
	protected Result<String> extractLocation(Tag eventTag) {
		return Payload.of(DEFAULT_LOCATION);
	}

	@Override
	protected Predicate<Tag> extractLocationFilter() {
		return null;
	}

	@Override
	protected Predicate<Tag> extractStartDateFilter() {
		return attributeEquals("itemprop","startDate");
	}

	@Override
	protected Predicate<Tag> extractStartTimeFilter() {
		return attributeEquals(CLASS,"event_time");
	}

	@Override
	protected List<String> extractTags(Tag eventTag) {
		return List.of("Kulturbahnhof","Jena","CosmicDawn");
	}

	@Override
	protected Predicate<Tag> extractTitleFilter() {
		return ofType("h1");
	}

	@Override
	protected Result<LocalDate> parseEndDate(String string) {
		return null;
	}

	@Override
	protected Result<LocalTime> parseEndTime(String string) {
		return null;
	}

	@Override
	protected Result<LocalDate> parseStartDate(String string) {
		return parseGermanDate(string);
	}

	@Override
	protected Result<LocalTime> parseStartTime(String string) {
		return parseGermanTime(string);
	}

		/**
	 * Die Kuba-Seite haut einen haufen Script mit raus, der dazu führt, dass die Tags nicht richtig geparsed werden.
	 * Also schneiden wir den kompletten header ab...
	 * @param inputStream eingehender InputStream, verpackt in Result
	 * @return ausgehender InputStream, verpackt in Result
	 */
	@Override
	protected Result<InputStream> preload(Result<InputStream> inputStream) {
		var opt = inputStream.optional();
		if (opt.isEmpty()) return transform(inputStream);
		try {
			var input = opt.get();
			var bos = new ByteArrayOutputStream();
			input.transferTo(bos);
			input.close();
			String code = bos.toString(UTF_8);
			var pos = code.indexOf("<body");
			return Payload.of(new ByteArrayInputStream(code.substring(pos).getBytes(UTF_8)));
		} catch (IOException e) {
			return error(e, "Failed to buffer data from %s", inputStream);
		}
	}

	@Override
	protected String programURL() {
		return baseUrl()+"/veranstaltungen/";
		//return "http://httpbin.org/headers";
	}
}
