/* © SRSoftware 2024 */
package de.srsoftware.cal.importer.jena;

import static de.srsoftware.cal.Util.parseGermanTime;
import static de.srsoftware.tools.Error.error;
import static de.srsoftware.tools.Result.transform;
import static de.srsoftware.tools.Tag.*;
import static de.srsoftware.tools.TagFilter.*;

import de.srsoftware.cal.BaseImporter;
import de.srsoftware.cal.api.Coords;
import de.srsoftware.tools.Payload;
import de.srsoftware.tools.Result;
import de.srsoftware.tools.Tag;
import java.security.NoSuchAlgorithmException;
import java.time.LocalDate;
import java.time.LocalTime;
import java.time.format.DateTimeFormatter;
import java.util.List;
import java.util.Objects;
import java.util.function.Predicate;

public class FHaus extends BaseImporter {

	// TODO: verwendet den gleichen Typ Kalender wie der Rosenkeller – evtl. kann man diese beiden in einer Abstrakten Superklasse zusammenführen

	public static final Coords DEFAULT_COORDS = new Coords(50.9293, 11.58228);
	public static final String DEFAULT_LOCATION = "F-Haus, Krautgasse 14, 07743 Jena";
	public FHaus() throws NoSuchAlgorithmException {
		super();
	}

	@Override
	protected String baseUrl() {
		return "https://www.f-haus.de";
	}

	@Override
	public String description() {
		return "Importer für Events des Jenaer F-Haus";
	}

	@Override
	protected Predicate<Tag> extractAttachmentsFilter() {
		return attributeEquals(ID,"tribe-events-content");
	}

	@Override
	protected Predicate<Tag> extractDescriptionFilter() {
		return attributeContains(CLASS,"single-event-description");
	}

	@Override
	protected Result<Coords> extractCoords(Tag eventTag) {
		return Payload.of(DEFAULT_COORDS);
	}

	@Override
	protected Predicate<Tag> extractEndDateFilter() {
		return attributeHas(CLASS,"tribe-events-abbr").and(attributeContains(CLASS,"tribe-events-end").and(attributeContains(CLASS,"date")));
	}

	@Override
	protected Predicate<Tag> extractEndTimeFilter() {
		return attributeHas(CLASS,"tribe-events-abbr").and(attributeContains(CLASS,"tribe-events-end").and(attributeContains(CLASS,"time")));
	}

	@Override
	protected Predicate<Tag> extractEventTagFilter() {
		return attributeEquals(ID,"tribe-events-content");
	}

	@Override
	protected Result<List<String>> extractEventUrls(Result<Tag> programPage) {
		var opt = programPage.optional();
		if (opt.isEmpty()) return transform(programPage);
		List<String> urls = opt.get().find(attributeEquals(CLASS,"tribe-events-calendar-list"))
				.stream().flatMap(tag -> tag.find(IS_ANCHOR).stream())
				.map(tag -> tag.get(HREF))
				.filter(Objects::nonNull)
				.filter(url -> url.contains("/event/"))
				.map(url -> url.contains("://") ? url : baseUrl()+url)
				.distinct()
				.toList();
		return Payload.of(urls);
	}

	@Override
	protected Predicate<Tag> extractLinksFilter() {
		return attributeContains(CLASS,"single-event-description");
	}

	@Override
	protected Result<String> extractLocation(Tag eventTag) {
		return Payload.of(DEFAULT_LOCATION);
	}

	@Override
	protected Predicate<Tag> extractLocationFilter() {
		return null;
	}

	protected Result<LocalDate> extractStartDate(Tag eventTag) {
		Result<Tag> startDateTag = extractStartDateTag(eventTag);
		var opt = startDateTag.optional();
		if (opt.isEmpty()) return transform(startDateTag);
		return parseStartDate(opt.get().get(TITLE));
	}

	protected Result<Tag> extractStartDateTag(Tag eventTag) {
		var list = eventTag.find(extractStartDateFilter());
		if (list.isEmpty()) {
			return error("Failed to find start date tag");
		}
		return Payload.of(list.getFirst());
	}

	@Override
	protected Predicate<Tag> extractStartDateFilter() {
		return attributeHas(CLASS,"tribe-events-abbr").and(attributeContains(CLASS,"tribe-events-start").and(attributeContains(CLASS,"date")));
	}

	@Override
	protected Predicate<Tag> extractStartTimeFilter() {
		return attributeHas(CLASS,"tribe-events-abbr").and(attributeContains(CLASS,"tribe-events-start").and(attributeContains(CLASS,"time")));
	}

	@Override
	protected List<String> extractTags(Tag eventTag) {
		return List.of("F-Haus", "Jena");
	}

	@Override
	protected Predicate<Tag> extractTitleFilter() {
		return ofType("h1");
	}

	@Override
	protected Result<LocalDate> parseEndDate(String string) {
		try {
			return Payload.of(LocalDate.parse(string, DateTimeFormatter.ISO_DATE));
		} catch (Exception e){
			return error(e,"Failed to parse date: %s",string);
		}
	}

	@Override
	protected Result<LocalTime> parseEndTime(String string) {
		return parseGermanTime(string);
	}

	@Override
	protected Result<LocalDate> parseStartDate(String string) {
		try {
			return Payload.of(LocalDate.parse(string, DateTimeFormatter.ISO_DATE));
		} catch (Exception e){
			return error(e,"Failed to parse date: %s",string);
		}
	}

	@Override
	protected Result<LocalTime> parseStartTime(String string) {
		return parseGermanTime(string);
	}

	@Override
	protected String programURL() {
		return baseUrl()+"/cms/events";
	}
}
