/* © SRSoftware 2024 */
package de.srsoftware.cal.importer.jena;

import static de.srsoftware.tools.Error.error;
import static de.srsoftware.tools.Result.transform;
import static de.srsoftware.tools.Tag.CLASS;
import static de.srsoftware.tools.Tag.ID;
import static de.srsoftware.tools.TagFilter.*;

import de.srsoftware.cal.SinglePageImporter;
import de.srsoftware.cal.Util;
import de.srsoftware.cal.api.Coords;
import de.srsoftware.tools.Payload;
import de.srsoftware.tools.Result;
import de.srsoftware.tools.Tag;
import java.security.NoSuchAlgorithmException;
import java.time.LocalDate;
import java.time.LocalTime;
import java.util.ArrayList;
import java.util.List;
import java.util.Objects;
import java.util.function.Predicate;
import java.util.regex.Pattern;

public class Psychochor extends SinglePageImporter {

	private static final Pattern DATE_TIME_PATTERN = Pattern.compile("(\\d{4}).*(\\d\\d+)\\W*([a-zA-Z]+)\\W*(\\d\\d?):(\\d\\d?)");
	private static final Pattern LATITUDE = Pattern.compile("!3d(-?\\d+\\.\\d{1,5})");
	private static final Pattern LONGITUDE = Pattern.compile("!2d(-?\\d+\\.\\d{1,5})");

	public Psychochor() throws NoSuchAlgorithmException {
		super();
	}

	@Override
	protected String baseUrl() {
		return "https://www.psycho-chor.de";
	}

	@Override
	public String description() {
		return "Importer für Events des Jenaer Psychochors";
	}

	@Override
	protected Predicate<Tag> extractAttachmentsFilter() {
		return attributeEquals("itemprop","description");
	}

	@Override
	protected Result<Coords> extractCoords(Tag eventTag) {
		var list = eventTag.find(ofType("iframe"));
		if (list.isEmpty()) return error("No iframe found");
		return list.stream().map(iframe -> iframe.get("data-src-cmplz"))
				.filter(Objects::nonNull)
				.map(this::parseCoords).findAny()
				.orElseGet(() -> error("No coordinates found!"));
	}

	private Result<Coords> parseCoords(String s) {
		// TODO: diese Koordinaten sind immer irgendwie daneben!?
		var latitude = LATITUDE.matcher(s);
		var longitude = LONGITUDE.matcher(s);
		if (latitude.find() && longitude.find()){
			double lat = Double.parseDouble(latitude.group(1));
			double lon = Double.parseDouble(longitude.group(1));
			return Payload.of(new Coords(lat,lon));
		}
		return error("Failed to parse coords from %s",s);
	}

	@Override
	protected Predicate<Tag> extractDescriptionFilter() {
		return attributeEquals("itemprop","description");
	}

	@Override
	protected Predicate<Tag> extractEndDateFilter() {
		return null;
	}

	@Override
	protected Predicate<Tag> extractEndTimeFilter() {
		return null;
	}

	@Override
	protected Predicate<Tag> extractEventTagFilter() {
		return null;
	}

	@Override
	protected Result<List<Tag>> extractEventTags(Result<Tag> tagResult) {
		var opt = tagResult.optional();
		if (opt.isEmpty()) return transform(tagResult);

		List<Tag> eventTags = opt.get().find(attributeEquals(ID, "evcal_list")).stream()
				.flatMap(tag -> tag.find(attributeStartsWith(ID, "event_")).stream())
				.toList();
		return Payload.of(eventTags);
	}

	@Override
	protected Predicate<Tag> extractLinksFilter() {
		return attributeEquals("itemprop","description");
	}

	@Override
	protected Predicate<Tag> extractLocationFilter() {
		return attributeContains(CLASS,"evcal_location");
	}

	@Override
	protected Predicate<Tag> extractStartDateFilter() {
		return attributeContains(CLASS,"evo_start");
	}

	@Override
	protected Predicate<Tag> extractStartTimeFilter() {
		return attributeContains(CLASS,"evo_start");
	}

	@Override
	protected List<String> extractTags(Tag eventTag) {
		var eventTags = new ArrayList<String>();
		eventTags.add("Psychochor");
		eventTags.add("Jena");
		eventTag.find(attributeEquals("data-filter","event_type")).stream().map(Tag::strip).forEach(eventTags::add);
		return eventTags;
	}

	@Override
	protected Predicate<Tag> extractTitleFilter() {
		return attributeHas(CLASS,"evcal_event_title");
	}

	@Override
	protected Result<LocalDate> parseEndDate(String string) {
		return null;
	}

	@Override
	protected Result<LocalTime> parseEndTime(String string) {
		return null;
	}

	@Override
	protected Result<LocalDate> parseStartDate(String string) {
		var matcher = DATE_TIME_PATTERN.matcher(string);
		if (matcher.find()){
			int year = Integer.parseInt(matcher.group(1));
			var res = Util.toNumericMonth(matcher.group(3));
			if (res.optional().isEmpty()) return transform(res);
			int month = res.optional().get();
			int day = Integer.parseInt(matcher.group(2));
			return Payload.of(LocalDate.of(year,month,day));
		}
		return error("Failed to parse date from %s",string);
	}

	@Override
	protected Result<LocalTime> parseStartTime(String string) {
		var matcher = DATE_TIME_PATTERN.matcher(string);
		if (matcher.find()){
			int hour = Integer.parseInt(matcher.group(4));
			int min = Integer.parseInt(matcher.group(5));
			return Payload.of(LocalTime.of(hour,min));
		}
		return error("Failed to parse date from %s",string);
	}

	@Override
	protected String programURL() {
		return baseUrl()+"/de/events";
	}
}
