/* © SRSoftware 2024 */
package de.srsoftware.cal.importer.jena;

import static de.srsoftware.cal.Util.parseGermanTime;
import static de.srsoftware.tools.Error.error;
import static de.srsoftware.tools.Result.transform;
import static de.srsoftware.tools.Tag.*;
import static de.srsoftware.tools.TagFilter.*;

import de.srsoftware.cal.BaseImporter;
import de.srsoftware.cal.api.Coords;
import de.srsoftware.tools.Payload;
import de.srsoftware.tools.Result;
import de.srsoftware.tools.Tag;
import java.security.NoSuchAlgorithmException;
import java.time.LocalDate;
import java.time.LocalTime;
import java.time.format.DateTimeFormatter;
import java.util.List;
import java.util.function.Predicate;

public class Rosenkeller extends BaseImporter {

	// TODO: verwendet den gleichen Typ Kalender wie das F-Haus – evtl. kann man diese beiden in einer Abstrakten Superklasse zusammenführen

	private static final String  BASE_URL	= "https://rosenkeller.org";
	private static final String  DEFAULT_LOCATION	= "Rosenkeller, Johannisstr. 13, 07743 Jena";
	private static final Coords COORDS = new Coords(50.92945, 11.58491);
	public Rosenkeller() throws NoSuchAlgorithmException {
		super();
	}

	@Override
	protected String baseUrl() {
		return BASE_URL;
	}

	@Override
	public String description() {
		return "Importiert Events des Studentenclubs „Rosenkeller“ in Jena";
	}

	@Override
	protected Predicate<Tag> extractAttachmentsFilter() {
		return attributeEquals(ID,"tribe-events-content");
	}

	@Override
	protected Predicate<Tag> extractDescriptionFilter() {
		return attributeContains(CLASS,"single-event-description");
	}

	@Override
	protected Result<Coords> extractCoords(Tag eventTag) {
		return Payload.of(COORDS);
	}

	protected Result<LocalDate> extractEndDate(Tag eventTag) {
		Result<Tag> endDateTag = extractEndDateTag(eventTag);
		var opt = endDateTag.optional();
		if (opt.isEmpty()) return transform(endDateTag);
		return parseEndDate(opt.get().get(TITLE));
	}

	@Override
	protected Predicate<Tag> extractEndDateFilter() {
		return attributeHas(CLASS,"tribe-events-abbr").and(attributeContains(CLASS,"tribe-events-end").and(attributeContains(CLASS,"date")));
	}

	@Override
	protected Predicate<Tag> extractEndTimeFilter() {
		return attributeHas(CLASS,"tribe-events-abbr").and(attributeContains(CLASS,"tribe-events-end").and(attributeContains(CLASS,"time")));
	}

	@Override
	protected Predicate<Tag> extractEventTagFilter() {
		return attributeEquals(ID,"tribe-events-content");
	}

	@Override
	protected Result<List<String>> extractEventUrls(Result<Tag> programPage) {
		if (programPage.optional().isEmpty()) return transform(programPage);
		List<String> list = programPage.optional()
				.get()	//
				.find(attributeStartsWith(ID, "event-"))
				.stream()
				.map(t -> t.find(attributeEquals(CLASS, "ect-event-url")))
				.flatMap(List::stream)
				.map(t -> t.get(HREF))
				.toList();
		return Payload.of(list);
	}

	@Override
	protected Predicate<Tag> extractLinksFilter() {
		return attributeContains(CLASS,"single-event-description");
	}

	@Override
	protected Result<String> extractLocation(Tag eventTag) {
		return Payload.of(DEFAULT_LOCATION);
	}

	@Override
	protected Predicate<Tag> extractLocationFilter() {
		return null;
	}

	protected Result<LocalDate> extractStartDate(Tag eventTag) {
		Result<Tag> startDateTag = extractStartDateTag(eventTag);
		var opt = startDateTag.optional();
		if (opt.isEmpty()) return transform(startDateTag);
		return parseStartDate(opt.get().get(TITLE));
	}

	protected Result<Tag> extractStartDateTag(Tag eventTag) {
		var list = eventTag.find(extractStartDateFilter());
		if (list.isEmpty()) {
			return error("Failed to find start date tag");
		}
		return Payload.of(list.getFirst());
	}

	@Override
	protected Predicate<Tag> extractStartDateFilter() {
		return attributeHas(CLASS,"tribe-events-abbr").and(attributeContains(CLASS,"tribe-events-start").and(attributeContains(CLASS,"date")));
	}

	@Override
	protected Predicate<Tag> extractStartTimeFilter() {
		return attributeHas(CLASS,"tribe-events-abbr").and(attributeContains(CLASS,"tribe-events-start").and(attributeContains(CLASS,"time")));
	}

	@Override
	protected List<String> extractTags(Tag eventTag) {
		return List.of("Rosenkeller", "Jena");
	}

	@Override
	protected Predicate<Tag> extractTitleFilter() {
		return ofType("h1");
	}

	@Override
	protected Result<LocalDate> parseEndDate(String string) {
		try {
			return Payload.of(LocalDate.parse(string, DateTimeFormatter.ISO_DATE));
		} catch (Exception e){
			return error(e,"Failed to parse date: %s",string);
		}
	}

	@Override
	protected Result<LocalTime> parseEndTime(String string) {
		return parseGermanTime(string);
	}

	@Override
	protected Result<LocalDate> parseStartDate(String string) {
		try {
			return Payload.of(LocalDate.parse(string, DateTimeFormatter.ISO_DATE));
		} catch (Exception e){
			return error(e,"Failed to parse date: %s",string);
		}
	}

	@Override
	protected Result<LocalTime> parseStartTime(String string) {
		return parseGermanTime(string);
	}

	@Override
	protected String programURL() {
		return baseUrl() + "/de/programm";
	}
}
